import zlib
from typing import Sequence

from clickhouse_connect.datatypes import registry
from clickhouse_connect.driver.common import read_leb128, read_leb128_str, write_leb128
from clickhouse_connect.driver.insert import InsertContext
from clickhouse_connect.driver.query import DataResult
from clickhouse_connect.driver.transform import DataTransform, QueryContext

block_size = 16384


class NativeTransform(DataTransform):
    # pylint: disable=too-many-locals
    def _transform_response(self, source: Sequence, context: QueryContext) -> DataResult:
        if not isinstance(source, memoryview):
            source = memoryview(source)
        loc = 0
        names = []
        col_types = []
        result = []
        total_size = len(source)
        block = 0
        use_none = context.use_none
        while loc < total_size:
            result_block = []
            num_cols, loc = read_leb128(source, loc)
            num_rows, loc = read_leb128(source, loc)
            for col_num in range(num_cols):
                name, loc = read_leb128_str(source, loc)
                if block == 0:
                    names.append(name)
                type_name, loc = read_leb128_str(source, loc)
                if block == 0:
                    col_type = registry.get_from_name(type_name)
                    col_types.append(col_type)
                else:
                    col_type = col_types[col_num]
                context.start_column(name, col_type)
                column, loc = col_type.read_native_column(source, loc, num_rows, use_none=use_none)
                result_block.append(column)
            if context.column_oriented:
                if block == 0:
                    result = [x if isinstance(x, list) else list(x) for x in result_block]
                else:
                    for base, added in zip(result, result_block):
                        base.extend(added)
            else:
                result.extend(list(zip(*result_block)))
            block += 1
        return DataResult(result, tuple(names), tuple(col_types), context.column_oriented)

    def _build_insert(self, context: InsertContext):
        if context.compression == 'gzip':
            compressor = GzipCompressor()
        else:
            compressor = NullCompressor()

        def chunk_gen():
            for x in context.next_block():
                output = bytearray()
                write_leb128(x.column_count, output)
                write_leb128(x.row_count, output)
                for col_name, col_type, data in zip(x.column_names, x.column_types, x.column_data):
                    write_leb128(len(col_name), output)
                    output += col_name.encode()
                    write_leb128(len(col_type.name), output)
                    output += col_type.name.encode()
                    try:
                        col_type.write_native_column(data, output)
                    except Exception as ex: # pylint: disable=broad-except
                        # This is hideous, but some low level serializations can fail while streaming
                        # the insert if the user has included bad data in the column.  We need to ensure that the
                        # insert fails (using garbage data) to avoid a partial insert, and use the context to
                        # propagate the correct exception to the user
                        context.insert_exception = ex
                        yield 'INTERNAL EXCEPTION WHILE SERIALIZING'.encode()
                        return
                yield compressor.compress_block(output)
            footer = compressor.complete()
            if footer:
                yield footer

        return chunk_gen()


class NullCompressor:
    @staticmethod
    def compress_block(block):
        return block

    def complete(self):
        pass


class GzipCompressor:
    def __init__(self, level: int = 6, wbits: int = 31):
        self.zlib_obj = zlib.compressobj(level=level, wbits=wbits)

    def compress_block(self, block):
        return self.zlib_obj.compress(block)

    def complete(self):
        return self.zlib_obj.flush()



from abc import ABC, abstractmethod
from typing import Sequence

from clickhouse_connect.driver.insert import InsertContext
from clickhouse_connect.driver.query import DataResult, QueryContext


_EMPTY_QUERY_CONTEXT = QueryContext()


class DataTransform(ABC):

    def parse_response(self, source: Sequence, context: QueryContext = _EMPTY_QUERY_CONTEXT) -> DataResult:
        """
        Decodes the ClickHouse byte buffer response into rows of native Python data
        :param source: A byte buffer or similar source
        :param context: The QueryContext to use in processing the response
        :return: DataResult -- data matrix, column names, column types
        """
        with context:
            return self._transform_response(source, context)

    def build_insert(self, context: InsertContext):
        """
        Encodes a dataset of Python sequences into a ClickHouse format using the specified insert context
        :param context InsertContext parameter object
        :return: generator of bytes like objects containing the dataset in the appropriate format
        """
        with context:
            return self._build_insert(context)

    @abstractmethod
    def _transform_response(self, source: Sequence, context: QueryContext) -> DataResult:
        pass

    @abstractmethod
    def _build_insert(self, context: InsertContext):
        pass



import socket
from ipaddress import IPv4Address, IPv6Address
from typing import Union, MutableSequence, Sequence

from clickhouse_connect.datatypes.base import ArrayType, ClickHouseType
from clickhouse_connect.driver.common import write_array, array_column

IPV4_V6_MASK = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff'
V6_NULL = bytes(b'\x00' * 16)
V4_NULL = IPv4Address(0)


# pylint: disable=protected-access
class IPv4(ArrayType):
    _array_type = 'I'
    valid_formats = 'string', 'native'

    @property
    def python_type(self):
        return str if self.read_format() == 'string' else IPv4Address

    @property
    def np_type(self):
        return 'U' if self.read_format() == 'string' else 'O'

    @property
    def python_null(self):
        return '' if self.read_format() == 'string' else V4_NULL

    def _read_native_binary(self, source: Sequence, loc: int, num_rows: int):
        if self.read_format() == 'string':
            return self._from_native_str(source, loc, num_rows)
        return self._from_native_ip(source, loc, num_rows)

    def _from_native_ip(self, source: Sequence, loc: int, num_rows: int):
        column, loc = array_column(self._array_type, source, loc, num_rows)
        fast_ip_v4 = IPv4Address.__new__
        new_col = []
        app = new_col.append
        for x in column:
            ipv4 = fast_ip_v4(IPv4Address)
            ipv4._ip = x
            app(ipv4)
        return new_col, loc

    def _from_native_str(self, source: Sequence, loc: int, num_rows: int, **_):
        column, loc = array_column(self._array_type, source, loc, num_rows)
        return [socket.inet_ntoa(x.to_bytes(4, 'big')) for x in column], loc

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        first = self._first_value(column)
        if isinstance(first, str):
            fixed = 24, 16, 8, 0
            # pylint: disable=consider-using-generator
            column = [(sum([int(b) << fixed[ix] for ix, b in enumerate(x.split('.'))])) if x else 0 for x in column]
        else:
            if self.nullable:
                column = [x._ip if x else 0 for x in column]
            else:
                column = [x._ip for x in column]
        write_array(self._array_type, column, dest)


# pylint: disable=protected-access
class IPv6(ClickHouseType):
    valid_formats = 'string', 'native'

    @property
    def python_type(self):
        return str if self.read_format() == 'string' else IPv6Address

    @property
    def np_type(self):
        return 'U' if self.read_format() == 'string' else 'O'

    @property
    def python_null(self):
        return '' if self.read_format() == 'string' else V6_NULL

    def _read_native_binary(self, source: Sequence, loc: int, num_rows: int):
        if self.read_format() == 'string':
            return self._read_native_str(source, loc, num_rows)
        return self._read_native_ip(source, loc, num_rows)

    @staticmethod
    def _read_native_ip(source: Sequence, loc: int, num_rows: int):
        fast_ip_v6 = IPv6Address.__new__
        fast_ip_v4 = IPv4Address.__new__
        with_scope_id = '_scope_id' in IPv6Address.__slots__
        new_col = []
        app = new_col.append
        ifb = int.from_bytes
        end = loc + (num_rows << 4)
        for ix in range(loc, end, 16):
            int_value = ifb(source[ix: ix + 16], 'big')
            if int_value >> 32 == 0xFFFF:
                ipv4 = fast_ip_v4(IPv4Address)
                ipv4._ip = int_value & 0xFFFFFFFF
                app(ipv4)
            else:
                ipv6 = fast_ip_v6(IPv6Address)
                ipv6._ip = int_value
                if with_scope_id:
                    ipv6._scope_id = None
                app(ipv6)
        return new_col, end

    @staticmethod
    def _read_native_str(source: Sequence, loc: int, num_rows: int):
        new_col = []
        app = new_col.append
        v4mask = IPV4_V6_MASK
        tov4 = socket.inet_ntoa
        tov6 = socket.inet_ntop
        af6 = socket.AF_INET6
        end = loc + (num_rows << 4)
        for ix in range(loc, end, 16):
            x = source[ix: ix + 16]
            if x[:12] == v4mask:
                app(tov4(x[12:]))
            else:
                app(tov6(af6, x))
        return new_col, end

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        v = V6_NULL
        first = self._first_value(column)
        v4mask = IPV4_V6_MASK
        af6 = socket.AF_INET6
        tov6 = socket.inet_pton
        if isinstance(first, str):
            for x in column:
                if x is None:
                    dest += v
                elif '.' in x:
                    dest += v4mask + bytes(int(b) for b in x.split('.'))
                else:
                    dest += tov6(af6, x)
        else:
            for x in column:
                if x is None:
                    dest += v
                else:
                    b = x.packed
                    dest += b if len(b) == 16 else (v4mask + b)



import array
from typing import Sequence, MutableSequence

from clickhouse_connect.json_impl import any_to_json
from clickhouse_connect.datatypes.base import ClickHouseType, TypeDef
from clickhouse_connect.driver.common import array_column, must_swap
from clickhouse_connect.datatypes.registry import get_from_name


class Array(ClickHouseType):
    __slots__ = ('element_type',)
    python_type = list

    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        self.element_type = get_from_name(type_def.values[0])
        self._name_suffix = f'({self.element_type.name})'

    def read_native_prefix(self, source: Sequence, loc: int):
        return self.element_type.read_native_prefix(source, loc)

    # pylint: disable=too-many-locals
    def read_native_data(self, source: Sequence, loc: int, num_rows: int, use_none: bool = True):
        final_type = self.element_type
        depth = 1
        while isinstance(final_type, Array):
            depth += 1
            final_type = final_type.element_type
        level_size = num_rows
        offset_sizes = []
        for _ in range(depth):
            level_offsets, loc = array_column('Q', source, loc, level_size)
            offset_sizes.append(level_offsets)
            level_size = level_offsets[-1] if level_offsets else 0
        if level_size:
            all_values, loc = final_type.read_native_data(source, loc, level_size, use_none)
        else:
            all_values = []
        column = all_values if isinstance(all_values, list) else list(all_values)
        for offset_range in reversed(offset_sizes):
            data = []
            last = 0
            for x in offset_range:
                data.append(column[last: x])
                last = x
            column = data
        return column, loc

    def write_native_prefix(self, dest: MutableSequence):
        self.element_type.write_native_prefix(dest)

    def write_native_data(self, column: Sequence, dest: MutableSequence):
        final_type = self.element_type
        depth = 1
        while isinstance(final_type, Array):
            depth += 1
            final_type = final_type.element_type
        for _ in range(depth):
            total = 0
            data = []
            offsets = array.array('Q')
            for x in column:
                total += len(x)
                offsets.append(total)
                data.extend(x)
            if must_swap:
                offsets.byteswap()
            dest += offsets.tobytes()
            column = data
        final_type.write_native_data(column, dest)


class Tuple(ClickHouseType):
    _slots = 'element_names', 'element_types'
    valid_formats = 'tuple', 'json', 'native'  # native is 'tuple' for unnamed tuples, and dict for named tuples

    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        self.element_names = type_def.keys
        self.element_types = [get_from_name(name) for name in type_def.values]
        if self.element_names:
            self._name_suffix = f"({', '.join(k + ' ' + str(v) for k, v in zip(type_def.keys, type_def.values))})"
        else:
            self._name_suffix = type_def.arg_str

    @property
    def python_type(self):
        if self.read_format() == 'tuple':
            return tuple
        if self.read_format() == 'json':
            return str
        return dict

    def read_native_prefix(self, source: Sequence, loc: int):
        for e_type in self.element_types:
            loc = e_type.read_native_prefix(source, loc)
        return loc

    def read_native_data(self, source: Sequence, loc: int, num_rows: int, use_none=True):
        columns = []
        e_names = self.element_names
        for e_type in self.element_types:
            column, loc = e_type.read_native_data(source, loc, num_rows, use_none)
            columns.append(column)
        if e_names and self.read_format() != 'tuple':
            dicts = [{} for _ in range(num_rows)]
            for ix, x in enumerate(dicts):
                for y, key in enumerate(e_names):
                    x[key] = columns[y][ix]
            if self.read_format() == 'json':
                to_json = any_to_json
                return [to_json(x) for x in dicts], loc
            return dicts, loc
        return tuple(zip(*columns)), loc

    def write_native_prefix(self, dest: MutableSequence):
        for e_type in self.element_types:
            e_type.write_native_prefix(dest)

    def write_native_data(self, column: Sequence, dest: MutableSequence):
        columns = list(zip(*column))
        for e_type, elem_column in zip(self.element_types, columns):
            e_type.write_native_data(elem_column, dest)


class Map(ClickHouseType):
    _slots = 'key_type', 'value_type'
    python_type = dict

    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        self.key_type = get_from_name(type_def.values[0])
        self.value_type = get_from_name(type_def.values[1])
        self._name_suffix = type_def.arg_str

    def read_native_prefix(self, source: Sequence, loc: int):
        loc = self.key_type.read_native_prefix(source, loc)
        loc = self.value_type.read_native_prefix(source, loc)
        return loc

    # pylint: disable=too-many-locals
    def read_native_data(self, source: Sequence, loc: int, num_rows: int, use_none=True):
        offsets, loc = array_column('Q', source, loc, num_rows)
        total_rows = offsets[-1]
        keys, loc = self.key_type.read_native_data(source, loc, total_rows, use_none)
        values, loc = self.value_type.read_native_data(source, loc, total_rows, use_none)
        all_pairs = tuple(zip(keys, values))
        column = []
        app = column.append
        last = 0
        for offset in offsets:
            app(dict(all_pairs[last: offset]))
            last = offset
        return column, loc

    def write_native_prefix(self, dest: MutableSequence):
        self.key_type.write_native_prefix(dest)
        self.value_type.write_native_prefix(dest)

    def write_native_data(self, column: Sequence, dest: MutableSequence):
        offsets = array.array('Q')
        keys = []
        values = []
        total = 0
        for v in column:
            total += len(v)
            offsets.append(total)
            keys.extend(v.keys())
            values.extend(v.values())
        if must_swap:
            offsets.byteswap()
        dest += offsets.tobytes()
        self.key_type.write_native_data(keys, dest)
        self.value_type.write_native_data(values, dest)


class Nested(ClickHouseType):
    __slots__ = 'tuple_array', 'element_names', 'element_types'
    python_type = Sequence[dict]

    def __init__(self, type_def):
        super().__init__(type_def)
        self.element_names = type_def.keys
        self.tuple_array = get_from_name(f"Array(Tuple({','.join(type_def.values)}))")
        self.element_types = self.tuple_array.element_type.element_types
        cols = [f'{x[0]} {x[1].name}' for x in zip(type_def.keys, self.element_types)]
        self._name_suffix = f"({', '.join(cols)})"

    def read_native_prefix(self, source: Sequence, loc: int):
        return self.tuple_array.read_native_prefix(source, loc)

    def read_native_data(self, source: Sequence, loc: int, num_rows: int, use_none: bool = True):
        keys = self.element_names
        data, loc = self.tuple_array.read_native_data(source, loc, num_rows, use_none)
        return [[dict(zip(keys, x)) for x in row] for row in data], loc

    def write_native_prefix(self, dest: MutableSequence):
        self.tuple_array.write_native_prefix(dest)

    def write_native_data(self, column: Sequence, dest: MutableSequence):
        keys = self.element_names
        data = [[tuple(sub_row[key] for key in keys) for sub_row in row] for row in column]
        self.tuple_array.write_native_data(data, dest)


class JSON(ClickHouseType):
    python_type = dict

    def write_native_prefix(self, dest: MutableSequence):
        dest.append(0x01)

    # pylint: disable=duplicate-code
    def write_native_data(self, column: Sequence, dest: MutableSequence):
        app = dest.append
        to_json = any_to_json
        for x in column:
            v = to_json(x)
            sz = len(v)
            while True:
                b = sz & 0x7f
                sz >>= 7
                if sz == 0:
                    app(b)
                    break
                app(0x80 | b)
            dest += v


class Object(JSON):
    python_type = dict

    def __init__(self, type_def):
        if type_def.values[0].lower() != "'json'":
            raise NotImplementedError('Only json Object type is currently supported')
        super().__init__(type_def)
        self._name_suffix = type_def.arg_str



import datetime
from typing import Iterable, Sequence, Optional, Any, Dict, NamedTuple, Generator

from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.driver.common import SliceView
from clickhouse_connect.driver.options import check_pandas, check_numpy
from clickhouse_connect.driver.exceptions import ProgrammingError

DEFAULT_BLOCK_SIZE = 16834


class InsertBlock(NamedTuple):
    column_count: int
    row_count: int
    column_names: Iterable[str]
    column_types: Iterable[ClickHouseType]
    column_data: Iterable[Sequence[Any]]


# pylint: disable=too-many-instance-attributes
class InsertContext:
    """
    Reusable Argument/parameter object for inserts.
    """
    # pylint: disable=too-many-arguments
    def __init__(self,
                 table: str,
                 column_names: Sequence[str],
                 column_types: Sequence[ClickHouseType],
                 data: Sequence[Sequence[Any]] = None,
                 column_oriented: bool = False,
                 settings: Optional[Dict[str, Any]] = None,
                 compression: Optional[str] = None,
                 block_size: int = DEFAULT_BLOCK_SIZE):
        self.table = table
        self.column_names = column_names
        self.column_types = column_types
        self.column_oriented = column_oriented
        self.settings = settings
        self.compression = compression
        self.block_size = block_size
        self.data = data
        self.insert_exception = None

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self.current_block = 0
        self.current_row = 0
        self.row_count = 0
        self.column_count = 0
        if data:
            if self.column_oriented:
                self._next_block_data = self._column_block_data
                self._block_columns = [SliceView(column) for column in data]
                self._block_rows = None
                self.column_count = len(data)
                self.row_count = len(data[0])
            else:
                self._next_block_data = self._row_block_data
                self._block_rows = data
                self._block_columns = None
                self.row_count = len(data)
                self.column_count = len(data[0])
        if self.row_count and self.column_count:
            if self.column_count != len(self.column_names):
                raise ProgrammingError('Insert data column count does not match column names')
            self._data = data
        else:
            self._data = None

    @property
    def df(self):
        raise NotImplementedError('DataFrame is read only')

    @df.setter
    def df(self, df):
        self.data = from_pandas_df(df, self.column_types)

    def next_block(self) -> Generator[InsertBlock, None, None]:
        while True:
            block_end = min(self.current_row + self.block_size, self.row_count)
            row_count = block_end - self.current_row
            if row_count <= 0:
                return
            self.current_block += 1
            data = self._next_block_data(self.current_row, block_end)
            yield InsertBlock(self.column_count, row_count, self.column_names, self.column_types, data)
            self.current_row = block_end

    def _column_block_data(self, block_start, block_end):
        return [col[block_start: block_end] for col in self._block_columns]

    def _row_block_data(self, block_start, block_end):
        column_slice = SliceView(self._block_rows[block_start: block_end])
        return tuple(zip(*column_slice))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def from_pandas_df(df, column_types: Sequence[ClickHouseType]):
    """
    Convert a pandas dataframe into Python native types for insert
    :param df: Pandas DataFrame for insert
    :param column_types: ClickHouse column types matching the data frame columns in order.  Used for transformation
    :return: Tuple of the dataframe column names and
    """
    np = check_numpy()
    pd = check_pandas()
    data = []
    for col_name, ch_type in zip(df.columns, column_types):
        df_col = df[col_name]
        d_type = str(df_col.dtype)
        if ch_type.python_type == int:
            if 'float' in d_type:
                df_col = df_col.round().astype(ch_type.base_type, copy=False)
            else:
                df_col = df_col.astype(ch_type.base_type, copy=False)
        elif ch_type.python_type in (datetime.datetime, datetime.date) and 'date' in d_type:
            data.append([None if pd.isnull(x) else pd.Timestamp.to_pydatetime(x) for x in df_col])
            continue
        if ch_type.nullable and ch_type.python_type != float:
            df_col.replace({np.nan: None}, inplace=True)
        data.append(df_col.tolist())
    return data
